#include <math.h>
#include <algorithm>
#include <stdlib.h>
#include <time.h>


/*
 * Game takes place on a rectangular area on the plane. The game logic and all
 * objects locations/dimensions are described by real numbers:
 * <0..1>X<0..ASPECT_RATIO>
 *
 * ASPECT_RATIO is calculated from window size: hight/width. When objects are
 * reentered all the dimensions are translated to real numbers from:
 * <0..WIDTH>X<0..HEIGHT> by multiplying through the window width.
 * */

#define AR ((double)app.h/app.w)


class geometry
{
public:
   struct P
   {
      P(): x(0), y(0) {}
      P(double x_, double y_): x(x_), y(y_) {}
      double x;
      double y;
   };
   static void intersec(double a1, double b1, double a2, double b2, P& p)
   {
      p.x = (b2-b1)/(a1-a2);
      p.y = p.x*a1+b1;
   }
   static void line(const P& l1, const P& l2, double& a, double& b)
   {
      a = (l1.y - l2.y)/(l1.x - l2.x);
      b =  l1.y - l1.x*a;
   }
   static inline double mod(double x, double y)
   {
      return sqrt(x*x+y*y);
   }

   static void perpendicular(const P& l1, const P& l2, const P& p, double& a, double& b)
   {
      double la = (l1.y - l2.y)/(l1.x - l2.x);
      a = -1/la;
      b = p.y - a*p.x;
   }

   static void closest_point_on_line(const P& l1, const P& l2, const P& p, P& c)
   {
      double perpend_a, perpend_b;
      double a, b;
      if (l1.y == l2.y)
      {
         c.y = l1.y;
         c.x = p.x;
         return;
      }
      if (l1.x == l2.x)
      {
         c.x = l1.x;
         c.y = p.y;
         return;
      }
      perpendicular(l1, l2, p, perpend_a, perpend_b);
      line(l1, l2, a, b);
      intersec(a, b, perpend_a, perpend_b, c);
   }

   /*
    * Time to colision between two objects and the edge.
    * */
   static double time_to_colision(const P& s, double dx, double dy, double xl, double yl, double size)
   {
      double xdistance = dx > 0 ?
         xl - s.x - size:
         -1*(s.x  - size);
      double tx = dx? xdistance/dx
         : 10000000;

      double ydistance = dy > 0 ?
         yl - s.y - size:
         -1*(s.y  - size);
      double ty = dy? ydistance/dy
         : 10000000;

      return std::min(tx, ty);
   }

   /*
    * Time to colision between two moving objects.
    * */
   static double time_to_colision(const P& s1, const P& s2, double dx1,
         double dx2, double dy1, double dy2, double size1, double size2)
   {
      // the equation is (size1+size2)^2 = (distance on x )^2 + (distance on y)^2,
      // which gives:
      //
      // (size1+size2)^2
      //    = (s1.x +dx1*t -s2.x -dx2*t)^2
      //    + (s1.y +dy1*t -s2.y -dy2*t)^2
      //
      // (size1+size2)^2
      //    = (s1.x-s2.x + t*(dx1-dx2))^2
      //    + (s1.y-s2.y + t*(dy1-dy2))^2
      //
      // (size1+size2)^2
      //    = (s1.x-s2.x)^2 + 2t*(s1.x-s2.x)*(dx1-dx2) +(dx1-dx2)^2*t^2
      //    + (s1.y-s2.y)^2 + 2t*(s1.y-s2.y)*(dy1-dy2) +(dy1-dy2)^2*t^2;
      //
      // written as equation on t:
      // t^2((dy1-dy2)*(dy1-dy2)+(dx1-dx2)*(dx1-dx2))
      //   + 2t((s1.x-s2.x)*(dx1-dx2)+(s1.y-s2.y)*(dy1-dy2))
      //   + (s1.x-s2.x)^2+(s1.y-s2.y)^2 - (size1+size2)^2
      //   = 0
      //
      // lets calculate it as a*t^2 + b*t +c = 0
      double a = (dy1-dy2)*(dy1-dy2) + (dx1-dx2)*(dx1-dx2);
      double b = 2*((s1.x-s2.x)*(dx1-dx2) + (s1.y-s2.y)*(dy1-dy2));
      double c = (s1.x-s2.x)*(s1.x-s2.x) + (s1.y-s2.y)*(s1.y-s2.y)
         - (size1+size2)*(size1+size2);

      // Now just to solve the equation. 
      double delta = b*b-4*a*c; 
      if (delta < 0)
         return 10000000; // No solution.

      // Return the lower positive solution.
      double result = (-b-sqrt(delta))/(2*a);
      if (result >= 0)
         return result;

      result = (-b+sqrt(delta))/(2*a);
      if (result >= 0)
         return result;

      return 10000000; // No solution.
   }
};

class Target
{
public:
   Target(App& application,
         double s, double pos_x, double pos_y,
         agg::rgba c):
      app(application), x(pos_x), y(pos_y), size(s), color(c), i(0) { }
   double x,y;
   double size;
   agg::rgba color;
   App& app;
   int i;

   void draw(agg::rasterizer_scanline_aa<>& ras,
         agg::scanline_u8& sl,
         renderer_base_type& rbase)
   {
      renderer_scanline_type ren_sl(rbase);
      ren_sl.color(color);
      agg::ellipse e1;
      agg::ellipse e2;
      ras.reset();
      double f   = (double)(i++%20)/20;
      double f_1 = 1-f;
      e1.init(app.w*x, app.w*y, app.w*size*f, app.w*size*f_1, 128);
      e2.init(app.w*x, app.w*y, app.w*size*f_1, app.w*size*f, 128);
      ras.add_path(e1);
      ras.add_path(e2);
      agg::render_scanlines(ras, sl, ren_sl);
   }
   void newLocation()
   {
      do {
         x = (double)rand()/RAND_MAX;
         y = (double)rand()/RAND_MAX*AR;
         // Don't put the target in any of the coroners or very close to the
         // edge.
      } while (false
            || x < 0.01 || x > 0.99
            || y < 0.01 || y > 0.99
            || (x < 0.03 && y < 0.03)
            || (x < 0.03 && y > 0.97*AR)
            || (x > 0.97 && y < 0.04)
            || (x > 0.97 && y > 0.97*AR)
            );
   }
};

class Ball;
class Target;
class Control
{
public:
   virtual void getAcc(double& acc_x, double& acc_y, Ball&, Target& t) = 0;
   virtual void draw(agg::rasterizer_scanline_aa<>& ras,
         agg::scanline_u8& sl,
         renderer_base_type& rbase) {};
};


class Ball
{
public:
   Ball(App& application,
         double s, double m, double pos_x, double pos_y,
         agg::rgba c, int img, Control* ctrl):
      app(application),
      size(s), mass(m), x(pos_x), y(pos_y), dx(0), dy(0),
      controller(ctrl), color(c), img_idx(img),
      force_field(0.3) { }
   double size;
   double mass;
   double x,y;   // location
   double dx,dy; // velocity
   double force_field;
   int    img_idx;
   agg::rgba color;
   App& app;
   Control* controller;

   void updateV(Target& target, double t)
   {
      double acc_x, acc_y;
      controller->getAcc(acc_x, acc_y, *this, target);
      dx += t*acc_x;
      dy += t*acc_y;
   }

   void draw(agg::rasterizer_scanline_aa<>& ras,
         agg::scanline_u8& sl,
         renderer_base_type& rbase)
   {
      renderer_scanline_type ren_sl(rbase);
      color.opacity(force_field);
      ren_sl.color(color);
      ras.reset();
      agg::ellipse e;
      e.init(app.w*x, app.w*y, app.w*size, app.w*size, 40);
      ras.add_path(e);
      agg::render_scanlines(ras, sl, ren_sl);
      //rbase.blend_from(agg::pixfmt_abgr32(app.rbuf_img(1)), 0, app.w*x-75, app.w*y-75);
      //return;

      agg::trans_affine img_mtx;
      double rotation = dx > 0 ? 0: agg::pi;
      rotation += (dx? atan(dy/dx) :0) - agg::pi/4;
      img_mtx *= agg::trans_affine_translation(-75, -75);
      img_mtx *= agg::trans_affine_rotation(rotation);
      img_mtx *= agg::trans_affine_translation(+75, +75);
      img_mtx *= agg::trans_affine_scaling(2.0*(app.w*size)/150);
      img_mtx *= agg::trans_affine_translation(app.w*(x-size), app.w*(y-size));
      img_mtx.invert();

      agg::span_allocator<agg::rgba8> sa;

      typedef agg::span_interpolator_linear<> interpolator_type;
      typedef agg::image_accessor_clip<pixfmt_type> img_source_type;
      typedef agg::span_image_filter_rgba_2x2<img_source_type,
              interpolator_type> span_gen_type;
      interpolator_type interpolator(img_mtx);

      pixfmt_type img_pixf(app.rbuf_img(img_idx));
      img_source_type img_src(img_pixf, agg::rgba_pre(0, 0.4, 0, 0.5));
      agg::image_filter<agg::image_filter_kaiser> filter;
      span_gen_type sg(img_src, interpolator, filter);

      ras.add_path(e);
      agg::render_scanlines_aa(ras, sl, rbase, sa, sg);
   }

   void move(double t)
   {
      x += t*dx;
      y += t*dy;

      if (force_field > 0.3)
         force_field -= 0.02;

      /* Against walls */
      if (dx < 0 && x-size <= 0)
         dx = -dx, force_field = 1.0;
      if (dx > 0 && x+size >= 1)
         dx = -dx, force_field = 1.0;
      if (dy < 0 && y-size <= 0)
         dy = -dy, force_field = 1.0;
      if (dy > 0 && y+size >= AR)
         dy = -dy, force_field = 1.0;
   }

   void collide(Ball& other)
   {
      double other_dx = other.dx;
      double other_dy = other.dy;

      geometry::P a(x, y);
      geometry::P b(other.x, other.y);
      geometry::P this_v(x+dx, y+dy);
      geometry::P c(0, 0);
      geometry::closest_point_on_line(a, b, this_v, c);

      dx = this_v.x - c.x;
      dy = this_v.y - c.y;
      other.dx = c.x - x;
      other.dy = c.y - y;

      geometry::P other_v(other.x + other_dx, other.y+other_dy);
      geometry::closest_point_on_line(a, b, other_v, c);
      other.dx += other_v.x - c.x;
      other.dy += other_v.y - c.y;
      dx += c.x - other.x;
      dy += c.y - other.y;

      force_field = 1.0;
      other.force_field = 1.0;
   }

   bool touchingTarget(Target& t)
   {
      return (t.x - x)*(t.x - x)+(t.y - y)*(t.y - y) < size*size;
   }
};


class AiControl: public Control 
{
public:
   void getAcc(double& acc_x, double& acc_y, Ball& b, Target& t)
   {
      bool goingSlow = b.dx*b.dx + b.dy*b.dy < 0.001*0.001;
      bool goingRight = geometry::time_to_colision(
            geometry::P(b.x, b.y),
            geometry::P(t.x, t.y),
            b.dx, 0, b.dy, 0, b.size, 0) < 5000;

      if (goingRight || goingSlow)
      {
         acc_x = 0.0003*(t.x - b.x); 
         acc_y = 0.0003*(t.y - b.y); 
         return;
      }
      acc_x = -0.2*b.dx; 
      acc_y = -0.2*b.dy; 
   }
};



class Control1: public Control
{
public:
   App& app;
   Control1(App& application):
      app(application) {}
   void getAcc(double& acc_x, double& acc_y, Ball& b, Target& t)
   {
      acc_x = 0; acc_y = 0;
      if (app.mouse_down)
      {
         acc_x = 0.002*(((double)app.mx)/app.w - b.x);
         acc_y = 0.002*(((double)app.my)/app.w - b.y);
      }
   }
};

class Control2: public Control
{
public:
   App& app;
   Control2(App& application, double size_, double x_, double y_):
      x(x_), y(y_), size(size_), app(application) {}
   void getAcc(double& acc_x, double& acc_y, Ball& b, Target& t)
   {
      acc_x = 0; acc_y = 0;
      double mx, my;
      bool active = false;

      if (app.mouse_down0 && d(app.mx0, app.my0) < 1.2*size)
      {
         mx = app.mx0;
         my = app.my0;
         active = true;
      }

      if (app.mouse_down1 && d(app.mx1, app.my1) < 1.2*size)
      {
         mx = app.mx1;
         my = app.my1;
         active = true;
      }
      
      if (active)
      {
         double dx = mx - x;
         double dy = my - y;
         double lt = 1;
         if (dx*dx+dy*dy > size*size)
         {
            lt = sqrt(size*size/(dx*dx+dy*dy));
         }
         acc_x = 0.004*(dx*lt);
         acc_y = 0.004*(dy*lt);
      }
   }
   void draw(agg::rasterizer_scanline_aa<>& ras,
         agg::scanline_u8& sl,
         renderer_base_type& rbase)
   {
      renderer_scanline_type ren_sl(rbase);
      ren_sl.color(lblue);
      agg::ellipse e1;
      agg::ellipse e2;
      ras.reset();
      e1.init(app.w*x, app.w*y, 4, 4, 30);
      e2.init(app.w*x, app.w*y, 60, 60, 30, true);
      ras.add_path(e1);
      ras.add_path(e2);
      agg::render_scanlines(ras, sl, ren_sl);
      double mx, my;
      bool active = false;

      if (app.mouse_down0 && d(app.mx0, app.my0) < 1.2*size)
      {
         mx = app.mx0;
         my = app.my0;
         active = true;
      }

      if (app.mouse_down1 && d(app.mx1, app.my1) < 1.2*size)
      {
         mx = app.mx1;
         my = app.my1;
         active = true;
      }
      
      if (active)
      {
         double dx = mx - x;
         double dy = my - y;
         double lt = 1;
         if (dx*dx+dy*dy > size*size)
         {
            lt = sqrt(size*size/(dx*dx+dy*dy));
         }
         double lx = x + dx*lt;
         double ly = y + dy*lt;
         ren_sl.color(red);
         agg::ellipse e3;
         e3.init(app.w*lx, app.w*ly, 10, 10, 30);
         ras.reset();
         ras.add_path(e3);
         agg::render_scanlines(ras, sl, ren_sl);
      }
   };
   double x, y, size;

private:
   double d(double mx, double my)
   {
      return sqrt((mx-x)*(mx-x)+(my-y)*(my-y));
   }
};

class Background
{
public:
   Background(App& application, agg::rgba c)
      : app(application), color(c), i(0)
   {
      on_resize();
   }
   void draw(
         agg::rasterizer_scanline_aa<>& ras,
         agg::scanline_u8& sl,
         renderer_base_type& rbase, bool blink) 
   {
      if (blink) {
         i = 10;
      }
      double f = i? (double)(i--)/10: 0;
      double f_1 = 1-f;
      agg::rgba c(
            f_1*color.r +f*lgray.r,
            f_1*color.g +f*lgray.g,
            f_1*color.b +f*lgray.b,
            f_1*color.a +f*lgray.a);
      rbase.clear(c);

      for (int i = 0; i < 50; i++)
      {
         renderer_scanline_type ren_sl(rbase);
         ren_sl.color(lgray);
         agg::ellipse e;
         ras.reset();
         double p = phase[i];
         e.init(s[i].x*app.w, s[i].y*app.w, 2+sin(p), 2+sin(p), 5);
         ras.add_path(e);
         agg::render_scanlines(ras, sl, ren_sl);
         phase[i] += 0.03;
      }
   }

   void on_resize() 
   {
      for (int i = 0; i < 50; i++)
      {
         s[i].x =      (double)rand()/RAND_MAX;
         s[i].y =   AR*(double)rand()/RAND_MAX;
         phase[i] =    (double)rand();
      }
   } 

   agg::rgba color;
   App& app;

   int i;
   geometry::P s[50];
   double phase[50];
};


class GameView : public View, public GameSettings
{
public:
   GameView(App& application): app(application),
      menu(80,  20, 180, 35,    "Menu", !flip_y),
      pause(220, 20, 320, 35,   "Pause", !flip_y),
      blink(false),
      background(app, dblue),
      aicontrol(),
      control1(app),
      controlP1(app, 0.2, 0.2, 0.3*AR),
      controlP2(app, 0.2, 0.8, 0.8*AR),
      p1ball(app, 0.06, 0.4, 0.9, 0.5, agg::rgba(0.9,0.9,0), rand()%5, &control1),
      p2ball(app, 0.06, 0.8, 0.1, 0.5, agg::rgba(0.9,0.0,0), rand()%5, &aicontrol),
      time_left(0),
      target(app, 0.03, 0.1, 0.6, agg::rgba(0, 0.7, 0.9))
   {
      menu.background_color(red);
      pause.background_color(red);
      add_ctrl(menu);
      add_ctrl(pause);
      frame_cnt = 0;
      app.start_timer();
   }

   void enter()
   {
      app.wait_mode(false);
      background.on_resize();
   }

   virtual void on_ctrl_change()
   {
      if (pause.status())
      {
         app.wait_mode(!app.wait_mode());
      }

      if (menu.status())
      {
         menu.status(false);
         app.changeView("menu");
      }
      app.force_redraw();
   }

   virtual void on_resize(int, int)
   {
      background.on_resize();
      controlP1.y = 0.3*AR;
      controlP2.y = 0.8*AR;
   }

   virtual void on_idle()
   {
      const int MAX_FPS = 35;
      int loop_time = app.elapsed_time();
#ifndef PROFILING
      if (loop_time < 1000/MAX_FPS)
      {
         usleep(1000*(1000/MAX_FPS-loop_time));
      }
      time_left -= (double)app.elapsed_time()/1000;
      app.start_timer();
#else
      if (frame_cnt++ == 1000)
      {
         printf("frames %.2f/s - 1000 in %.2fms\n",
               (double)1000*frame_cnt/app.elapsed_time(),
               app.elapsed_time());
         app.start_timer();
         frame_cnt = 0;
      }
#endif
      update();
      app.force_redraw();
   }

   virtual void on_mouse_button_up(int x, int y, unsigned flags)
   {
      if (menu.on_mouse_button_up(x, y)
            || pause.on_mouse_button_up(x, y))
      {
         on_ctrl_change();
         return;
      }
      app.mouse_down = false;
   }

   virtual void on_mouse_button_down(int x, int y, unsigned flags)
   {
      if (menu.on_mouse_button_down(x, y)
            || pause.on_mouse_button_down(x, y))
      {
         on_ctrl_change();
         return;
      }
      app.mouse_down = true;
      app.mx = x;
      app.my = y;
   }

   virtual void on_mouse_move(int x, int y, unsigned flags)
   {
      app.mx = x;
      app.my = y;
   }

   void on_touch_event(float x, float y, float dx, float dy,
         int id, bool down)
   {
      if (id == 0)
      {
         if (!down)
            app.mouse_down0 = false;
         else
         {
            app.mouse_down0 = true;
            app.mx0 = x;
            app.my0 = y*AR;
         }
      }
      if (id == 1)
      {
         if (!down)
            app.mouse_down1 = false;
         else
         {
            app.mouse_down1 = true;
            app.mx1 = x;
            app.my1 = y*AR;
         }
      }
   }

   virtual void on_draw()
   {
      pixfmt_type pf(app.rbuf_window());;
      renderer_base_type rbase(pf);
      agg::rasterizer_scanline_aa<> ras;
      agg::scanline_u8 sl;
      ras.reset();
      background.draw(ras, sl, rbase, blink);
      blink = false;

      p1ball.controller->draw(ras, sl, rbase);
      p2ball.controller->draw(ras, sl, rbase);
      p1ball.draw(ras, sl, rbase);
      p2ball.draw(ras, sl, rbase);
      target.draw(ras, sl, rbase);

      char scores[100];
      sprintf(scores, "%.1f, p1:%d, p2:%d", time_left, app.p1Socore, app.p2Socore);
      app.draw_text(10, 60, 15, scores);

      agg::render_ctrl(ras, sl, rbase, pause);
      agg::render_ctrl(ras, sl, rbase, menu);
   }

   void setController(Player p, CtrlType t)
   {
      ballByPlayer(p).controller = ctrlByType(t, p);
   }

private:
    App& app;
    agg::button_ctrl<agg::rgba8> pause;
    agg::button_ctrl<agg::rgba8> menu;
    Ball p1ball;
    Ball p2ball;
    Target target;
    Control1 control1;
    AiControl aicontrol;
    Control2 controlP1;
    Control2 controlP2;
    Background background;
    bool blink;
    int frame_cnt;
    double time_left;

    Ball& ballByPlayer(Player p)
    {
       switch(p)
       {
          case PLAYER1:
             return p1ball;
          case PLAYER2:
             return p2ball;
       }

    }
    Control* ctrlByType(CtrlType t, Player p)
    {
       switch(t)
       {
          case TOUCH_TO_GO:
             return &control1;
          case VIRT_JOYSTICK:
             if (p == PLAYER1)
                return &controlP1;
             else
                return &controlP2;
          case AI_CONTROLLER:
             return &aicontrol;
       }
    }

    void update()
    {
       double acc_x = 0;
       double acc_y = 0;

       double t1 = geometry::time_to_colision(
             geometry::P(p1ball.x, p1ball.y),
             p1ball.dx, p1ball.dy,
             1, AR,
             p1ball.size);
       double t2 = geometry::time_to_colision(
             geometry::P(p2ball.x, p2ball.y),
             p2ball.dx, p2ball.dy,
             1, AR,
             p2ball.size);
       double t3 = geometry::time_to_colision(
             geometry::P(p1ball.x, p1ball.y),
             geometry::P(p2ball.x, p2ball.y),
             p1ball.dx, p2ball.dx,
             p1ball.dy, p2ball.dy,
             p1ball.size, p2ball.size);
       double t = std::min(1.0, std::min(std::min(t1, t2), t3));

       p1ball.move(t);
       p2ball.move(t);
       if (t == t3)
       {
          p1ball.collide(p2ball);
          p1ball.move(0.1*t);
          p2ball.move(0.1*t);
          if (app.sound)
             app.play_sound(1, geometry::mod(p1ball.dx-p2ball.dx,p1ball.dy-p2ball.dy)*1500);
       }
       p1ball.updateV(target, 1.0);
       p2ball.updateV(target, 1.0);


       if (p1ball.touchingTarget(target))
       {
          app.p1Socore++;
          blink = true;
          target.newLocation();
          if (app.sound)
             app.play_sound(2, 40);
       }
       if (p2ball.touchingTarget(target))
       {
          app.p2Socore++;
          blink = true;
          target.newLocation();
          if (app.sound)
             app.play_sound(2, 40);
       }

       /* Change music track from time to time...*/
       if (blink && app.music && (app.p1Socore + app.p2Socore) % 10 == 0)
       {
         app.play_music(rand()%4, 40);
       }

       if (newGame)
       {
          newGame = false;
          p1ball.dx = 0;
          p1ball.dy = 0;
          p1ball.x = 0.9;
          p1ball.y = 0.5;

          p2ball.dx = 0;
          p2ball.dy = 0;
          p2ball.x = 0.1;
          p2ball.y = 0.5;

          time_left = 60;
       }

       if (time_left <= 0)
       {
         app.changeView("game_end");
       }
    }
};
